'''
Created on Dec 19, 2016

@author: klaehn
'''
from time import time, sleep
from communiate import Mqtt
from data_buffer import DataBuffer
from light_sensor.light_sensor import light_sensor
from gate.gate_handler import GateHandler
from power_sensor import PowerSensor

STATE_INIT = "init"
STATE_OPENED = "open"
STATE_CLOSED = "close"
STATE_OPENING = "opening"
STATE_CLOSING = "closing"
STATE_ERROR = "error"

LIGHT_READ_DELAY_S = 30
LIGHT_CONSECUTIVE_READS = 10
LIGHT_LX_THRESHOLD = {"open":5, "close":0}

MQTT_HOST = "gitlab"
MQTT_TOPIC = "outdoor/chickenhouse/gate"

LIGHT_SENSOR_I2C_BUS = 1
LIGHT_SENSOR_I2C_ADDRESS = 0x23

POWER_SENSOR_I2C_BUS = 1
POWER_SENSOR_I2C_ADDRESS = 0x40
POWER_CONSECUTIVE_READS = 1000

MAX_ENGINE_POWER = {"up":310, "down":290}
MAX_GATE_RUNTIME = {"open":300, "close":250}


class GateState(object):
    def __init__(self):
        self.__state_handler = {STATE_INIT:self.__init_handler, \
                                STATE_OPENED:self.__opened_handler, \
                                STATE_CLOSED:self.__closed_handler, \
                                STATE_OPENING:self.__opening_handler, \
                                STATE_CLOSING:self.__closing_handler, \
                                STATE_ERROR:self.__error_handler}
        self.__next_state = "init"
        self.__last_state = "error"

        self.__light_sensor = light_sensor(LIGHT_SENSOR_I2C_BUS, \
                                           LIGHT_SENSOR_I2C_ADDRESS)
        self.__light_data = DataBuffer(LIGHT_CONSECUTIVE_READS)
        self.__comserver = Mqtt(MQTT_HOST)
        self.__gate_handler = GateHandler()
        self.__power_sensor = PowerSensor(POWER_SENSOR_I2C_BUS, \
                                          POWER_SENSOR_I2C_ADDRESS)
        self.__power_data = DataBuffer(POWER_CONSECUTIVE_READS)
        self.__gate_move_timeout = 0
        self.__light_read_timeout = 0
        self.__error_count = 0

    def poll(self):
        current_time = time()
        if current_time >= self.__light_read_timeout:
            self.__light_read_timeout = current_time + LIGHT_READ_DELAY_S
            self.__light_data.push(self.__light_sensor.read())
        self.__state_handler[self.__next_state](self.__light_data.average())

    def __update_state(self, new_state):
        self.__last_state = self.__next_state
        self.__next_state = new_state


    def __is_transition(self):
        if self.__last_state != self.__next_state:
            return True
        return False


    def __init_handler(self, light_avg):
        '''
        In init we don't know anything neither about gate state nor about
        light. So first we try to reach STATE_CLOSED.
        '''
        #pylint: disable=unused-argument
        self.__comserver.connect()
        self.__comserver.transmit(MQTT_TOPIC, str(time()) + \
                                  " gate gard initiated")
        self.__comserver.disconnect()
        self.__update_state(STATE_CLOSING)


    def __opened_handler(self, light_avg):
        next_state = self.__next_state
        if self.__is_transition():
            self.__gate_handler.close()
            sleep(5)
            self.__gate_handler.stop()
            self.__comserver.transmit(MQTT_TOPIC, str(time()) + " Opened " + \
                                      str(self.__power_data.average()) + " mW")
            self.__power_data.clear()

        if light_avg <= LIGHT_LX_THRESHOLD["close"]:
            next_state = STATE_CLOSING

        self.__update_state(next_state)


    def __closed_handler(self, light_avg):
        next_state = self.__next_state
        if self.__is_transition():
            self.__gate_handler.open()
            sleep(5)
            self.__gate_handler.stop()
            self.__comserver.transmit(MQTT_TOPIC, str(time()) + " Closed " + \
                                      str(self.__power_data.average()) + " mW")
            self.__power_data.clear()

        if light_avg > LIGHT_LX_THRESHOLD["open"]:
            next_state = STATE_OPENING

        self.__update_state(next_state)


    def __opening_handler(self, light_avg):
        next_state = self.__next_state
        if self.__is_transition():
            self.__gate_handler.open()
            self.__gate_move_timeout = time() + MAX_GATE_RUNTIME["open"]
            self.__comserver.transmit(MQTT_TOPIC, str(time()) + \
                                      " Opening " + str(light_avg) + " lx")

        if time() > self.__gate_move_timeout:
            next_state = STATE_ERROR
        else:
            self.__power_data.push(self.__power_sensor.power_mw())
            current_avg = self.__power_data.average()
            if current_avg != None:
                if current_avg > MAX_ENGINE_POWER["up"]:
                    next_state = STATE_OPENED

        self.__update_state(next_state)

    def __closing_handler(self, light_avg):
        next_state = self.__next_state
        if self.__is_transition():
            self.__gate_handler.close()
            self.__gate_move_timeout = time() + MAX_GATE_RUNTIME["close"]
            self.__comserver.transmit(MQTT_TOPIC, str(time()) + \
                                      " Closing " + str(light_avg) + " lx")

        if time() > self.__gate_move_timeout:
            next_state = STATE_ERROR
        else:
            self.__power_data.push(self.__power_sensor.power_mw())
            current_avg = self.__power_data.average()
            if current_avg != None:
                if current_avg > MAX_ENGINE_POWER["down"]:
                    next_state = STATE_CLOSED

        self.__update_state(next_state)


    def __error_handler(self, light_avg):
        #pylint: disable=unused-argument
        if self.__is_transition():
            self.__gate_handler.stop()
            self.__comserver.transmit(MQTT_TOPIC, str(time()) + \
                                      " Error handler!!!")
        if self.__error_count >= 5:
            self.__update_state(STATE_INIT)
            self.__error_count += 1
        else:
            self.__comserver.transmit(MQTT_TOPIC, str(time()) + \
                                      "Error Handler - couldn't reinitiate " + \
                                      "system. Bye bye...")
